unit MultiLanguage;

{
  EasyMenus
  Copyright 2005 Murariu Alexandru
  muralexandru@yahoo.com
  All rights reserved.
  For conditions of distribution and use, see LICENSE.TXT.

  Multilanguage Component
  SOME FEATURES:
   - Reads/Writes any string property of any TComponent
     (like Self.Caption, Edit1.Text, etc).
   - Reads/Writes any TStrings property of any TComponent
     (like TMemo.Lines or TCombobox.Items, etc).
   - Saves/Loads To/From Text File (INI).
   - There is even an event that warns if the loaded language is older than the application
     (so you can say to the user to check your website for an updated language file).
   - Modify the IgnoreProperties to specify which properties are saved/loaded and which not.
   - More... (See below)

}

interface

uses
  Windows, Messages, SysUtils, Classes, TypInfo, IniFiles,
  Forms, Menus, Dialogs;

type

  TOnWriteCompData = procedure (Sender : TObject; Comp : TComponent; PropName : string;
                    var Value : string; var DoWrite : boolean) of Object;
  TOnReadCompData = procedure (Sender : TObject; Comp : TComponent; PropName : string;
                    var Value : string; var DoChange : boolean) of Object;

  TMultiLanguage = class(TComponent)
  private
    FINI : TMemIniFile;
    FFileStrings,
    FIgnoreProperties,
    FLanguageFiles : THashedStringList;
    FFileName : string;
    FOnWriteProp : TOnWriteCompData;
    FOnReadProp : TOnReadCompData;
    FBefSave,
    FBefLoad,
    FAfterSave,
    FAfterLoad,
    FOnOldLangVersionLoaded : TNotifyEvent;
    FAuthorEmail,
    FAuthor,
    FLanguageName,
    FLanguageVersion,
    FAppVersion,
    FFileExt,
    FLanguagesDirectory : string;
    FAutoSaveExistentForms : boolean;
    FLangMenuItem : TMenuItem;
    FUseBitmaps : boolean;
    FOnRefreshed: TNotifyEvent;
    procedure SetStrings(Value : TStrings);
    function GetStrings : TStrings;
    procedure LangHeader(Save : boolean);
    procedure CheckVersion;
    procedure SetIgnore(Value : TStrings);
    function GetIgnore : TStrings;
    function GetFLanguageFiles : TStrings;
    procedure OnItemClick(Sender : TObject);
    function GetLanguageName(FileName : string) : string;
    procedure UpdateMenuDisplay;
    procedure SetFLangMenuItem(Value : TMenuItem);
    procedure SETFUseBitmaps(Value : boolean);
    procedure SETFLanguagesDirectory(Value : string);
    function GetFLanguageNames : TStrings;
    procedure RemoveItems;
    procedure PopulateLanguageList;
    procedure RefreshLangMenuItem;
    { Private declarations }
  protected
    procedure Notification(AComponent: TComponent;
    Operation: TOperation); override;
    { Protected declarations }
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure Lang_WriteComponent(AComp, AOwner : TComponent); //Write CompData (once) - Call Save After
    procedure Lang_ReadComponent(AComp, AOwner : TComponent);  //Read CompData (from language file) - Call Load After
    procedure Lang_ReadRootComponent(AComp : TComponent; ReadOwnedComponents : boolean = True); //Write CompData Owner and Owned (once) - Call Save After
    procedure Lang_WriteRootComponent(AComp : TComponent; WriteOwnedComponents : boolean = True); //Read CompData Owner and Owned (from language file) - Call Load After
    procedure WriteScreenForms;
    procedure ReadScreenForms;
    procedure Save(ToFile : string);
    procedure Load(FromFile : string);
    procedure Refresh;
    property LanguageFiles : TStrings read GetFLanguageFiles;
    property LanguageNames : TStrings read GetFLanguageNames;
    property FileName : string read FFileName;
    { Public declarations }
  published
    property AutoSaveExistentForms : boolean read FAutoSaveExistentForms write FAutoSaveExistentForms;
    property AppVersion : string read FAppVersion write FAppVersion;
    property Author : string read FAuthor write FAuthor;
    property AuthorEmail : string read FAuthorEmail write FAuthorEmail;
    property LanguageName : string read FLanguageName write FLanguageName;
    property LanguageVersion : string read FLanguageVersion write FLanguageVersion;
    property LanguageStrings : TStrings read GetStrings write SetStrings;
    property OnWriteProperty : TOnWriteCompData read FOnWriteProp write FOnWriteProp;
    property OnReadProperty : TOnReadCompData read FOnReadProp write FOnReadProp;
    property OnBeforeSave : TNotifyEvent read FBefSave write FBefSave;
    property OnBeforeLoad : TNotifyEvent read FBefLoad write FBefLoad;
    property OnAfterSave : TNotifyEvent read FAfterSave write FAfterSave;
    property OnAfterLoad : TNotifyEvent read FAfterLoad write FAfterLoad;
    property OnOldLangVersionLoaded : TNotifyEvent read FOnOldLangVersionLoaded write FOnOldLangVersionLoaded;
    property IgnoreProperties : TStrings read GetIgnore write SetIgnore;
    property FileExtension : string read FFileExt write FFileExt;
    property LanguagesDirectory : string read FLanguagesDirectory write SETFLanguagesDirectory;
    property LangMenuItem : TMenuItem read FLangMenuItem write SetFLangMenuItem;
    property UseLanguageBitmaps : boolean read FUseBitmaps write SETFUseBitmaps;
    property OnRefreshed : TNotifyEvent read FOnRefreshed write FOnRefreshed;
    { Published declarations }
  end;

  const SECTIONGEN = 'General';
        DefaultIgnoreArray : array[0..3] of string = ('Name', 'ImeName', 'HelpKeyword', 'HelpFile');

  resourcestring
  SUNTITLED = '<Untitled Language>';

implementation
uses EM_Common;

  {$R NoFlag.res}

procedure TMultiLanguage.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  case Operation of
    opInsert:
      begin

      end;
    opRemove:
      begin
      if LangMenuItem = AComponent then LangMenuItem := nil;
      end;
  end;
end;

function TMultiLanguage.GetLanguageName(FileName : string) : string;
var INI : TIniFile;
begin
try
INI := TIniFile.Create(FileName);
result := INI.ReadString(SECTIONGEN, 'LanguageName', '');
if result = '' then result := SUNTITLED;
INI.Free;
except
FreeAndNil(INI);
end;
end;

procedure TMultiLanguage.Refresh;
begin
if csDesigning in Self.ComponentState then exit;
PopulateLanguageList;
if Assigned(OnRefreshed) then OnRefreshed(Self);
end;

procedure TMultiLanguage.OnItemClick(Sender : TObject);
begin
Load(FLanguageFiles.Strings[(Sender as TMenuItem).Tag - 1]);
end;

procedure TMultiLanguage.UpdateMenuDisplay;
var I : integer;
begin
if csDesigning in Self.ComponentState then exit;
if not Assigned(FLangMenuItem) then exit;
for I := 0 to FLangMenuItem.Count - 1 do
    if FLangMenuItem.Items[I].Tag <> 0 then
       begin
       FLangMenuItem.Items[I].Checked := SameText(FLanguageFiles.Strings[LangMenuItem.Items[I].Tag - 1], FFileName);
       end;
end;

procedure TMultiLanguage.SETFLanguagesDirectory(Value : string);
begin
FLanguagesDirectory := Value;
Refresh;
end;

procedure TMultiLanguage.RemoveItems;
var I : Integer;
label 1;
begin
if not Assigned(FLangMenuItem) then exit;
1:for I := 0 to FLangMenuItem.Count - 1 do
    if FLangMenuItem.Items[I].Tag <> 0 then
               begin
               FLangMenuItem.Items[i].Free;
               goto 1;
               end;
end;

procedure TMultiLanguage.RefreshLangMenuItem;
var I : integer;
    MI : TMenuItem;
    P : Integer;
begin
if csDesigning in Self.ComponentState then exit;
RemoveItems;
if not Assigned(FLangMenuItem) then exit;
P := FLangMenuItem.Count - 1;
if P < 1 then P := 1;
for I := 0 to FLanguageFiles.Count - 1 do
    begin
    MI := TMenuItem.Create(nil);
    MI.Caption := GetLanguageName(FLanguageFiles.Strings[I]);
    MI.Tag := I + 1;
    if FUseBitmaps then
       begin
       if FileExists(ChangeFileExt(FLanguageFiles.Strings[I],'.bmp')) then MI.Bitmap.LoadFromFile(ChangeFileExt(FLanguageFiles.Strings[I],'.bmp'))
                                                                      else MI.Bitmap.LoadFromResourceName(HInstance, 'NOFLAG'); // Don't localize
       end;
    MI.OnClick := OnItemClick;
    FLangMenuItem.Insert(P + I - 1,MI);
    end;
UpdateMenuDisplay;
end;

procedure TMultiLanguage.PopulateLanguageList;
var Path, FPath : string;
    sr: TSearchRec;
    FileAttrs: Integer;
begin
if csDesigning in Self.ComponentState then exit;
FPath := FLanguagesDirectory;
if FPath = '' then FPath := GetCurrentDir;
Path := IncludeTrailingPathDelimiter(FPath); //Works on Windows and Linux
FPath := Path + '*.' + FFileExt;
FLanguageFiles.Clear;
FileAttrs := faAnyFile - faDirectory;
   if FindFirst(FPath, FileAttrs, sr) = 0 then
    begin
      repeat
        if (sr.Attr and FileAttrs) = sr.Attr then
        begin
        FLanguageFiles.Add(Path + sr.Name);
        end;
      until FindNext(sr) <> 0;
      FindClose(sr);
    end;
RefreshLangMenuItem;
end;

procedure TMultiLanguage.SetFLangMenuItem(Value : TMenuItem);
begin
  if FLangMenuItem <> Value then
  begin
    FLangMenuItem := Value;
    if Value <> nil then Value.FreeNotification(Self);    
    RefreshLangMenuItem;
  end;
end;

function TMultiLanguage.GetFLanguageNames : TStrings;
var i : integer;
begin
result := TStringList.Create;
for i := 0 to FLanguageFiles.Count - 1 do
    begin                       
    result.Add(GetLanguageName(FLanguageFiles.Strings[i]));
    end;
end;

procedure TMultiLanguage.SETFUseBitmaps(Value : boolean);
begin
FUseBitmaps := Value;
RefreshLangMenuItem;
end;

constructor TMultiLanguage.Create(AOwner : TComponent);
var i : integer;
begin
inherited Create(AOwner);
FINI := TMemIniFile.Create('');
FINI.WriteString(SECTIONGEN, 'LanguageVersion', '1.0'); //Dummy
FFileStrings := THashedStringList.Create;
FIgnoreProperties := THashedStringList.Create;
FLanguageFiles := THashedStringList.Create;
for i := Low(DefaultIgnoreArray) to High(DefaultIgnoreArray) do FIgnoreProperties.Add(DefaultIgnoreArray[i]);
FAutoSaveExistentForms := True;
FFileExt := 'lng';
FUseBitmaps := False;
FAppVersion := '1.0';
FLanguageVersion := '1.0';
Refresh;
end;

destructor TMultiLanguage.Destroy;
begin
RemoveItems;
if Assigned(FINI) then FINI.Free;
if Assigned(FFileStrings) then FFileStrings.Free;
if Assigned(FIgnoreProperties) then FIgnoreProperties.Free;
if Assigned(FLanguageFiles) then FLanguageFiles.Free;
inherited Destroy;
end;

procedure TMultiLanguage.CheckVersion;
begin
if IsVersionOlder(LanguageVersion,AppVersion) then
   if Assigned(OnOldLangVersionLoaded) then OnOldLangVersionLoaded(Self);
end;

procedure TMultiLanguage.LangHeader(Save : boolean);
begin
if Save then
   begin
   FINI.WriteString(SECTIONGEN, 'Author', FAuthor);
   FINI.WriteString(SECTIONGEN, 'AuthorEmail', FAuthorEmail);
   FINI.WriteString(SECTIONGEN, 'LanguageName', FLanguageName);
   FINI.WriteString(SECTIONGEN, 'LanguageVersion', FLanguageVersion);
   end
   else
   begin
   FAuthor := FINI.ReadString(SECTIONGEN, 'Author', '<Unknown>');
   FAuthorEmail := FINI.ReadString(SECTIONGEN, 'AuthorEmail', '');
   FLanguageName := FINI.ReadString(SECTIONGEN, 'LanguageName', ExtractFileName(FFileName));
   FLanguageVersion := FINI.ReadString(SECTIONGEN, 'LanguageVersion', '1.0');
   end;
end;

procedure TMultiLanguage.Save(ToFile : string);
begin
if Assigned(OnBeforeSave) then OnBeforeSave(Self);
FFileStrings.Clear;
LangHeader(True);
if FAutoSaveExistentForms then WriteScreenForms;
FINI.GetStrings(FFileStrings);
FFileStrings.SaveToFile(ToFile);
FFileName := ToFile;
FFileStrings.Clear;
if Assigned(OnAfterSave) then OnAfterSave(Self);
end;

procedure TMultiLanguage.Load(FromFile : string);
begin
if Assigned(OnBeforeLoad) then OnBeforeLoad(Self);
FFileStrings.Clear;
FFileStrings.LoadFromFile(FromFile);
FINI.Clear;
FINI.SetStrings(FFileStrings);
FFileName := FromFile;
LangHeader(False);
FFileStrings.Clear;
ReadScreenForms;
if Assigned(OnAfterLoad) then OnAfterLoad(Self);
CheckVersion;
UpdateMenuDisplay;
end;

procedure TMultiLanguage.SetIgnore(Value : TStrings);
begin
FIgnoreProperties.Assign(Value);
end;

function TMultiLanguage.GetIgnore : TStrings;
begin
result := TStringList.Create;
result.Assign(FIgnoreProperties);
end;

function TMultiLanguage.GetFLanguageFiles : TStrings;
begin
result := TStringList.Create;
result.Assign(FLanguageFiles);
end;

procedure TMultiLanguage.SetStrings(Value : TStrings);
begin
FINI.SetStrings(Value);
LangHeader(False);
end;

function TMultiLanguage.GetStrings : TStrings;
begin
result := TStringList.Create;
LangHeader(True);
FINI.GetStrings(result);
end;

procedure TMultiLanguage.WriteScreenForms;
var i : integer;
begin
for i := 0 to Screen.FormCount - 1 do
    Lang_WriteRootComponent(Screen.Forms[i]);
end;

procedure TMultiLanguage.ReadScreenForms;
var i : integer;
begin
for i := 0 to Screen.FormCount - 1 do
    Lang_ReadRootComponent(Screen.Forms[i]);
end;

procedure TMultiLanguage.Lang_ReadRootComponent(AComp : TComponent; ReadOwnedComponents : boolean = True);
var i : integer;
begin
Lang_ReadComponent(AComp,AComp);
if ReadOwnedComponents then
   for i := 0 to AComp.ComponentCount - 1 do
       Lang_ReadComponent(Acomp.Components[i],AComp);
end;

procedure TMultiLanguage.Lang_WriteRootComponent(AComp : TComponent; WriteOwnedComponents : boolean = True);
var i : integer;
begin
Lang_WriteComponent(AComp,Acomp);
if WriteOwnedComponents then
   for i := 0 to AComp.ComponentCount - 1 do
       Lang_WriteComponent(Acomp.Components[i],AComp);
end;

procedure TMultiLanguage.Lang_ReadComponent(AComp, AOwner : TComponent);
var P : PPropList;
    i, j : integer;
    Value : string;
    DoRead : boolean;
    Ow : String;
    ST : TStrings;
begin
if (AComp.Name = '') or (AOwner.Name = '') then exit; //Can't process unnamed components
if AComp is TMultiLanguage then exit;
if AOwner <> nil then Ow := AOwner.Name
                 else Ow := '';
if (AOwner is TForm) then
   begin
   if (AOwner as TForm).FormStyle = fsMDIChild then Ow := AOwner.ClassName; // On MDI forms save in the ClassName section because they are created dinamicaly at runtime with different names.
   end;
for i:=0 to GetPropList(AComp,p)-1 do
      begin
      if (not Assigned(p[i].GetProc)) or (not Assigned(p[i].SetProc)) then continue;
      if P[i].PropType^.Kind in [tkLString, tkString] then //Only want strings
         begin
         if FIgnoreProperties.IndexOf(p[i].Name) <> -1 then continue;
         if p[i].Name = '' then continue;
         if SameText(p[i].Name, 'Name') then continue; //Don't read component name
         Value := FINI.ReadString(Ow, AComp.Name + '.' + p[i].Name, '');
         if Value = '' then continue;
         DoRead := True; //By default change string
         if Assigned(OnReadProperty) then OnReadProperty(Self, AComp, p[i].Name, Value, DoRead); //Call event(can cancel changing or change the value)
         if DoRead then SetStrProp(AComp, p[i].name, Value);
         end
         else
         begin
      if P[i].PropType^.Kind in [tkClass] then //Only want TStrings
         if (GetObjectProp(AComp,P[i].Name) is TStrings) then
            begin
            ST := (GetObjectProp(AComp,P[i].Name) as TStrings); //Get the string
            for j := 0 to ST.Count - 1 do
                begin
                DoRead := True; //By default write string
                Value := FINI.ReadString(Ow, AComp.Name + '.' + p[i].Name + '.Strings['+ inttostr(j) +']', '');
                if Value = '' then continue;
                if Assigned(OnReadProperty) then OnReadProperty(Self, AComp, p[i].Name + '.Strings['+ inttostr(j) +']', Value, DoRead); //Call event(can cancel writing or change the value)
                if DoRead then ST.Strings[j] := Value;
                end;
            end;
         end;
      end;
end;

procedure TMultiLanguage.Lang_WriteComponent(AComp, AOwner : TComponent);
var P : PPropList;
    i, j : integer;
    Value : string;
    DoWrite : boolean;
    Ow : string;
    ST : TStrings;
begin
if (AComp.Name = '') or (AOwner.Name = '') then exit; //Can't process unnamed components
if AComp is TMultiLanguage then exit;
if AOwner <> nil then Ow := AOwner.Name
                 else Ow := '';
if (AOwner is TForm) then
   begin
   if (AOwner as TForm).FormStyle = fsMDIChild then Ow := AOwner.ClassName; // On MDI forms save in the ClassName section because they are created dinamicaly at runtime with different names.
   end;
for i:=0 to GetPropList(AComp,p)-1 do
      begin
      if (not Assigned(p[i].GetProc)) or (not Assigned(p[i].SetProc)) then continue;
      if P[i].PropType^.Kind in [tkLString, tkString] then //Only want strings
         begin
         if FIgnoreProperties.IndexOf(p[i].Name) <> -1 then continue;
         if p[i].Name = '' then continue;
         if SameText(p[i].Name, 'Name') then continue; //Don't write component name
         Value := GetStrProp(AComp, p[i].Name); //Get the string
         DoWrite := True; //By default write string
         if Assigned(OnWriteProperty) then OnWriteProperty(Self, AComp, p[i].Name, Value, DoWrite); //Call event(can cancel writing or change the value)
         if DoWrite then FINI.WriteString(Ow, AComp.Name + '.' + p[i].Name, Value);
         end
         else
         begin
      if P[i].PropType^.Kind in [tkClass] then //Only want TStrings
         if (GetObjectProp(AComp,P[i].Name) is TStrings) then
            begin
            ST := (GetObjectProp(AComp,P[i].Name) as TStrings); //Get the string
            for j := 0 to ST.Count - 1 do
                begin
                DoWrite := True; //By default write string
                Value := ST.Strings[j];
                if Assigned(OnWriteProperty) then OnWriteProperty(Self, AComp, p[i].Name + '.Strings['+ inttostr(j) +']', Value, DoWrite); //Call event(can cancel writing or change the value)
                if DoWrite then FINI.WriteString(Ow, AComp.Name + '.' + p[i].Name + '.Strings['+ inttostr(j) +']', Value);
                end;
            end;
         end;
      end;
end;

end.
 